-- SPDX-FileCopyrightText: 2025 Nathan Ringo <me@remexre.com>
--
-- SPDX-License-Identifier: MIT OR Apache-2.0

local namespace = vim.api.nvim_create_namespace("nvim-sylvan")

if vim.g.nvim_sylvan_cmd == nil then
	vim.g.nvim_sylvan_cmd = {"sylvan", "--json", "std"}
end

local M = {}

local function find_buffer(path)
	for buffer = 1, vim.fn.bufnr("$") do
		if vim.fn.buflisted(buffer) == 1 then
			if vim.api.nvim_buf_get_name(buffer) == path then
				return buffer
			end
		end
	end
	return nil
end

local function translate_diagnostic(diagnostic)
  return {
    lnum = diagnostic.start.line - 1,
    col = diagnostic.start.column,
    end_lnum = diagnostic["end"].line - 1,
    end_col = diagnostic["end"].column,
    code = diagnostic.code,
    message = diagnostic.kind .. "[" .. diagnostic.code .. "]: " .. diagnostic.message,
    severity = ({
      error = vim.diagnostic.severity.ERROR,
      warn = vim.diagnostic.severity.WARN,
      debug = vim.diagnostic.severity.INFO,
    })[diagnostic.kind] or vim.diagnostic.severity.ERROR
  }
end

M.diagnostic_opts = {}
local function add_diagnostics(buffer, diagnostics)
	local vim_diagnostics = {}
	for i, diagnostic in pairs(diagnostics) do
		vim_diagnostics[i] = translate_diagnostic(diagnostic)
	end
	vim.diagnostic.set(namespace, buffer, vim_diagnostics, M.diagnostic_opts)
end

local function update_diagnostics_from_command(obj)
	if obj.signal ~= 0 then
		error(vim.inspect(cmd) .. " failed to execute: " .. vim.inspect(obj))
	end
	vim.api.nvim_err_write(obj.stderr)

	local all_diagnostics = {}
	local beginning_of_line = 1
	while beginning_of_line < string.len(obj.stdout) do
		local end_of_line = string.find(obj.stdout, "\n", beginning_of_line)
		local line = string.sub(obj.stdout, beginning_of_line, end_of_line)
		local diagnostic = vim.json.decode(line)
		beginning_of_line = end_of_line + 1

		if diagnostic.path ~= nil then
			if all_diagnostics[diagnostic.path] == nil then
				all_diagnostics[diagnostic.path] = {}
			end
			table.insert(all_diagnostics[diagnostic.path], diagnostic)
		end
	end

	for path, diagnostics in pairs(all_diagnostics) do
		local buffer = find_buffer(path)
		if buffer ~= nil then
			add_diagnostics(buffer, diagnostics)
		end
	end
end

function M.clear_diagnostics()
  vim.diagnostic.reset(namespace, nil)
end

function M.update_diagnostics(cmd)
  M.clear_diagnostics()
	vim.system(cmd, { text = true }, function (obj)
    vim.schedule(function()
      update_diagnostics_from_command(obj)
    end)
  end)
end

function M.create_autocmd()
  vim.api.nvim_create_autocmd("BufWritePre", {
    pattern = {"*.syl"},
    callback = function(ev)
      M.update_diagnostics(vim.g.nvim_sylvan_cmd)
    end
  })
end

return M

-- vim: set cc=80 et sw=2 ts=2 :
